package Cm;

import java.util.*;
import java.io.*;
import java.net.*;
import java.awt.*;

class CmServerInfosHandler extends Handler
{
  public CmServerInfosHandler ()
      {
        super ("CmServerInfos");
      }

  public void run (Message message, String sender)
      {
        String version = message.get_text (); // should be v7
        String name    = message.get_text ();
        String host    = message.get_text ();
        int port       = message.get_int ();
        int multiple   = message.get_int ();
        String owner   = message.get_text ();

        Connect connect = Server.get_connection (sender);
        connect.set_infos (name, host, port, multiple, owner);
      }
}

class NSPortHandler extends Handler
{
  public NSPortHandler ()
      {
        super ("NSPort");
      }

  public void run (Message message, String sender)
      {
        String name = message.get_text ();
        if (name == null)
          {
            System.out.println ("Port handler> no name");
            return;
          }

        int port = message.get_int ();
        String new_name = message.get_text ();
        if (new_name == null)
          {
            System.out.print ("Port handler> no new name");
          }

        if (new_name != null)
          {
            name = new_name;
            Server.set_server_name (name);
          }

        if (port == 0)
          {
            System.out.println ("Cm> Name already used.");
            return;
          }

        ServerSocket socket = null;

        try
          {
            socket = new ServerSocket (port);
          }
        catch (Exception e)
          {
            System.out.print ("NSPortHandler> Error creating the server");
            System.out.print (" port=");
            System.out.println (port);
          }

        if (socket != null)
          {
            InternalServer server = new InternalServer (name, socket);
          }
      }
}

class NSNewPortHandler extends Handler
{
  public NSNewPortHandler ()
      {
        super ("NSNewPort");
      }

  public void run (Message message, String sender)
      {
      }
}

class NSAddressHandler extends Handler
{
  public NSAddressHandler ()
      {
        super ("NSAddress");
      }

  public void run (Message message, String sender)
      {
          //System.out.println ("NSAddress handler");

        int transaction = 0;
        String name = message.get_text();
        String host = message.get_text();
        if (host != null)
          {
            int port = message.get_int ();
            if (port != 0)
              {
                int multiple = message.get_int ();
                String owner = message.get_text ();
                transaction = message.get_int ();

                Connect connect = Server.find_connection (name);
                if (connect == null)
                  {
                    Socket socket = null;
                    try
                      {
                        socket = new Socket (host, port);
                      }
                    catch (Exception e)
                      {
                        System.out.print ("Cannot create socket " + host + port);
                        System.out.println (" for "+ name);
                      }

                    Connect c = new Connect (name, socket);
                    c.set_infos (name, host, port, multiple, owner);
                    Server.send_server_infos (socket);
                  }
                else
                  {
                      // connection already there
                    connect.set_infos (name, host, port, multiple, owner);
                  }
              }
          }
        else
          {
            transaction = message.get_int ();
          }

        Transaction.terminate (transaction);
      }
}

class NSWakeupHandler extends Handler
{
  public NSWakeupHandler ()
      {
        super ("NSWakeup");
      }

  public void run (Message message, String sender)
      {
        System.out.println ("Wakeup handler");
      }
}

class RequestComment extends Handler
{
  public RequestComment ()
      {
        super ("CmRequestComment");
      }

  public void run (Message message, String sender)
      {
        Message answer = new Message ("CmComment");
        answer.put_text ("alive");
        answer.send (sender);
      }
}

class InternalServer extends Thread
{
  public InternalServer (String name, ServerSocket socket)
      {
        _name = name;
        _socket = socket;

        start ();
      }

  public void run ()
      {
        while (true)
          {
            try
              {
                Socket socket = _socket.accept ();

                if (socket != null)
                  {
                    _server_number++;
                    String name = "__temp__" + String.valueOf (_server_number);

                    Connect connect = new Connect (name, socket);
                  }
              }
            catch (Exception e)
              {
                System.out.println ("run> accept error");
              }
          }
      }

  private
  String _name;
  ServerSocket _socket;
  static int _server_number = 0;
}

public class Server
{
  public static void set_domain (String domain)
      {
        _domain_infos = get_domain_infos (domain);

        if (_domain_infos != null)
          {
            StringTokenizer splitter = new StringTokenizer (_domain_infos);

            String name = splitter.nextToken ();
            _name_server_host = splitter.nextToken ();
            _name_server_port = Integer.parseInt (splitter.nextToken ());
          }

        new CmServerInfosHandler ();
        new NSPortHandler ();
        new NSNewPortHandler ();
        new NSAddressHandler ();
        new NSWakeupHandler ();
        new RequestComment ();
      }

  public static void open (String name)
      {
        String domain = System.getProperty ("CMDOMAIN");
        if (domain == null) domain = "VirgoLAL";
          //System.out.print ("open_server> domain=");
          //System.out.println (domain);
        set_domain (domain);
        Socket s = connect_to_name_server ();
        _multiple = false;
        set_server_name (name);
        send_server_infos (s);
      }

  public static void open_multiple (String name)
      {
        String domain = System.getProperty ("CMDOMAIN");
        if (domain == null) domain = "VirgoLAL";
          //System.out.print ("open_multiple_server> domain=");
          //System.out.println (domain);
        set_domain (domain);
        Socket s = connect_to_name_server ();
        _multiple = true;
        set_server_name (name);
        send_server_infos (s);
      }

  private static String get_domain_infos (String domain_name)
      {
        String result = null;

        try
          {
            URL url = new URL ("http://www.lal.in2p3.fr/SI/Cm/CmDomains");
            InputStream stream = url.openStream ();
            BufferedReader file = new BufferedReader (new InputStreamReader (stream));
            while (true)
              {
                try
                  {
                    String line = file.readLine ();
                    if (line == null) break;

                    StringTokenizer splitter;

                    splitter = new StringTokenizer (line);
                  
                    String name;

                    name = splitter.nextToken ();
                    if (name.compareTo (domain_name) == 0) 
                      {
                        result = line;
                      }
                  }
                catch (Exception e)
                  {
                    break;
                  }
              }
          }
        catch (Exception e)
          {
            System.out.println ("get_domain_infos>Error");
          }

        return (result);
      }

  private static Socket connect_to_name_server ()
      {
        Connect connect = get_connection ("CmNameServer");

        Socket socket = null;

        if (connect != null)
          {
            socket = connect.get_socket ();
          }

        if (socket == null)
          {
            try
              {
                if (_name_server_host != null)
                  {
                    socket = new Socket (_name_server_host,
                                         _name_server_port);
                  }
              }
            catch (Exception e)
              {
                System.out.println ("connect_to_name_server> Error");
              }
          }

        if (socket != null)
          {
            if (connect == null)
              {
                connect = new Connect ("CmNameServer", socket);
              }
            else
              {
                connect.set_socket (socket);
              }
          }

        return (socket);
      }

  public static void send_server_infos (Socket s)
      {
        Message m = new Message ();

        m.set_type ("CmServerInfos");

        m.put_text ("v7");               // Cm version
        m.put_text (_server_name);       // proposed name

        String host;

        try
          {
            InetAddress add = InetAddress.getLocalHost ();
            host = add.getHostName ();
              //host = add.getHostAddress ();
          }
        catch (Exception e)
          {
            System.out.println ("no local host");
            host = "asc";
          }

          // System.out.println ("local host : " + host);

        m.put_text (host);               // host
        m.put_int  (0);                  // port #
        m.put_int  (_multiple ? 1 : 0);  // is multiple

        String owner;
        try
          {
            owner = System.getProperty ("user.name");
          }
        catch (Exception e)
          {
            owner = "xxx";
          }

        m.put_text (owner);

        m.send (s);
      }

  public static void add_connection (String name, Connect connect)
      {
        if (_connections == null)
          {
            _connections = new Hashtable ();
          }

        _connections.put (name, connect);
      }

  public static void remove_connection (String name)
      {
        if (_connections == null) return;

        _connections.remove (name);
      }

  public static Connect get_connection (String name)
      {
        Connect connect;

        if (_connections == null)
          {
            // System.out.println ("no connections");
            return (null);
          }

        connect = (Connect) _connections.get (name);

        if (name == "CmNameServer")
          {
	    return (connect);
          }


        if (connect == null)
          {
            int transaction = Transaction.open ("Cm> get address");

            Message m = new Message ("NSGetAddress");

            Socket sock = connect_to_name_server ();

            m.put_text (name);
            m.put_int (0);
            m.put_int ((int) transaction);

            if (m.send (sock))
              {
                Transaction.do_wait (transaction);
                connect = (Connect) _connections.get (name);
              }

            Transaction.close (transaction);
            return (connect);
          }

        return (connect);
      }

  public static Connect find_connection (String name)
      {
	return ((Connect) _connections.get (name) );
      }

  public static void set_server_name (String name)
      {
        _server_name = name;
      }

  public static void set_server_port (int port)
      {
        _server_port = port;
      }

  private static String    _domain_infos;
  private static String    _name_server_host;
  private static int       _name_server_port;
  private static String    _server_name;
  private static int       _server_port;
  private static boolean   _multiple;
  private static Hashtable _connections;
}
